# mHPoT trainer
# 
# training on natural image patches using Fast Persistent Contrastive 
# Divergence or Contrastive Divergence. Samples are drawn from the model
# using Hamiltonian Monte Carlo.
# Input patches are whitened (i.e. re-scaled PCA coefficients of leading 
# eigenvectors).
# The whitening transform is: S^-.5 * U' where U collects in its columns 
# the leading eigenvectors, and S is a diagonal matrix with the corresponsding 
# eigenvalues.
# 
# Free energy: 
# F(x) = sum_j gamma log(1 + .5 sum_f P_fj (sum_i C_fi x_i)^2) +
#        + .5 sum_i x_i^2 - sum_k log(1+exp(sum_i W_ki x_i +b_k)) - sum_i b^v_i x_i
# Probability distribution: p(x) proportional to exp(-F(x))
# 
# This corresponds to an energy function defined over two sets of latent variables.
# The latent variables associated to the first term (the covariance latent variables) are
# gamma distributed, while those associated with the last term (the mean latent variables)
# are Bernoulli distributed.
#
# In the code below, we call:
#  gamma -> coeff
#  P -> FH
#  C -> VF
#  W -> w_mean
#  b -> bias_mean
#  b^v -> bias_vis
#
# For more information aobut the model, see eq. 1 through 6 in:
# M. Ranzato, V. Mnih, G. Hinton, "Generating More Realistic Images Using Gated MRF's". 
# Advances in Neural Information Processing Systems 23 (NIPS 2010), Vancouver, 2010
# available at http://www.cs.toronto.edu/~ranzato/publications/ranzato_nips2010.pdf
#
# Marc'Aurelio Ranzato
# 17 December 2010
# ranzato@cs.toronto.edu

import sys
from pylab import *
import numpy as np
import cudamat as cmt
from scipy.io import loadmat, savemat
import gpu_lock # put here you locking system package, if any
from ConfigParser import *

def compute_energy_mHPoT(data,vel,energy,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t6,tt0,tt1,tt2,feat,featsq,feat_mean):
    ''' This function computes the value of the hamiltonian at each data point in the batch
    The output is a vector of scalars (as many as data points in the batch)
    Return values: F(x) + .5 vel^2 for each data point'''
    # potential
    # covariance contribution
    cmt.dot(VF.T, data, target = feat) # HxV VxP -> HxP 
    feat.mult(feat, target = featsq)   # HxP
    cmt.dot(FH.T,featsq, target = t1) # OxH HxP ->  OxP, t1=(VF*x)^2*FH 
    t1.mult(0.5)
    t1.add(1) # OxP 1+.5(VF*x)^2*FH
    cmt.log(t1)
    t1.mult_by_col(coeff) # coeff log(1+.5(VF*x)^2*FH) 
    t1.sum(axis = 0, target = energy)
    # mean contribution
    cmt.dot(w_mean.T, data, target = feat_mean) # KxV VxP -> KxP 
    feat_mean.add_col_vec(bias_mean) # KxP
    feat_mean.greater_than(tt0,target = tt1) # avoid numerical overflow while computing log(1+exp())
    tt1.mult(-1.,target=tt2)
    tt2.add(1.)    
    tt1.mult(feat_mean)
    feat_mean.mult(tt2)
    cmt.exp(feat_mean) 
    feat_mean.add(1.)
    cmt.log(feat_mean)
    feat_mean.mult(tt2)
    feat_mean.add(tt1)
    feat_mean.mult(-1.) # - log(1+exp(W X + b))
    energy.add_sums(feat_mean, axis=0)
    # add quadratic term: .5 x^2
    data.mult(data,target = t6) 
    t6.mult(.5)
    energy.add_sums(t6, axis=0)    
    # visible bias term: - b x
    data.mult_by_col(bias_vis, target = t6)
    t6.mult(-1.) # DxP
    energy.add_sums(t6,  axis=0) # 1xP
    # kinetic: .5 vel^2
    vel.mult(vel, target = t6)
    energy.add_sums(t6, axis = 0, mult = .5)

def compute_gradient_mHPoT(data,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t2,t3,feat,featsq,feat_mean,gradient):
    '''  This function computes the derivatives of F(x) w.r.t. the input (for each input data point) '''
    # compute derivatives through covariance term in the energy function
    cmt.dot(VF.T, data, target = feat) # HxP 
    feat.mult(feat, target = featsq)   # HxP
    cmt.dot(FH.T,featsq, target = t1) # OxP, t1=(VF*x)^2*FH 
    t1.mult(0.5)
    t1.add(1) # OxP 1+.5(VF*x)^2*FH
    t1.reciprocal(target = t2) # OxP 1/(1+.5(VF*x)^2*FH)
    t2.mult_by_col(coeff) # OxP coeff/(1+.5(VF*x)^2*FH)
    cmt.dot(FH,t2, target = t3) # HxP
    t3.mult(feat) # HxP (coeff FH(VF*x))/(1+.5(VF*x)^2*FH)
    cmt.dot(VF,t3, target = gradient) # DxP gradient of first term
    # add visible bias term
    gradient.add_col_mult(bias_vis, -1.)
    # add MEAN contribution to gradient
    cmt.dot(w_mean.T, data, target = feat_mean) # HxP 
    feat_mean.add_col_vec(bias_mean) # HxP
    feat_mean.apply_sigmoid() # expected value of posterior: h
    feat_mean.mult(-1.)
    gradient.add_dot(w_mean, feat_mean) # add - W h
    # add derivative of quadratic term: x
    gradient.add(data)

def draw_HMC_samples(data,negdata,vel,gradient,new_energy,old_energy,VF,FH,coeff,bias_vis,w_mean,bias_mean,hmc_step,hmc_step_nr,hmc_ave_rej,hmc_target_ave_rej,t1,t2,t3,t4,t5,t6,t7,tt0,tt1,tt2,thresh,feat,featsq,batch_size,feat_mean):
    ''' Hybrid Monte Carlo sampler
    This is used to draw samples from the model (defined in terms of an energy function over the input pixels)'''
    vel.fill_with_randn()
    negdata.assign(data)
    compute_energy_mHPoT(negdata,vel,old_energy,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t6,tt0,tt1,tt2,feat,featsq,feat_mean)
    compute_gradient_mHPoT(negdata,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t2,t3,feat,featsq,feat_mean,gradient)
    # half step
    vel.add_mult(gradient, -0.5*hmc_step)
    negdata.add_mult(vel,hmc_step)
    # full leap-frog steps
    for ss in range(hmc_step_nr - 1):
        ## re-evaluate the gradient
        compute_gradient_mHPoT(negdata,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t2,t3,feat,featsq,feat_mean,gradient)
        # update variables
        vel.add_mult(gradient, -hmc_step)
        negdata.add_mult(vel,hmc_step)
    # final half-step
    compute_gradient_mHPoT(negdata,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t2,t3,feat,featsq,feat_mean,gradient)
    vel.add_mult(gradient, -0.5*hmc_step)
    # compute new energy
    compute_energy_mHPoT(negdata,vel,new_energy,VF,FH,coeff,w_mean,bias_mean,bias_vis,t1,t6,tt0,tt1,tt2,feat,featsq,feat_mean)
    # rejecton
    old_energy.subtract(new_energy, target = thresh)
    cmt.exp(thresh)
    t4.fill_with_rand()
    t4.less_than(thresh)
    # update negdata and rejection rate
    t4.mult(-1)
    t4.add(1) # now 1's detect rejections
    t4.sum(axis = 1, target = t5)
    t5.copy_to_host()
    rej = t5.numpy_array[0,0]/batch_size
    data.mult_by_row(t4, target = t6)
    negdata.mult_by_row(t4, target = t7)
    negdata.subtract(t7)
    negdata.add(t6)
    hmc_ave_rej = 0.9*hmc_ave_rej + 0.1*rej # running average rejection rate
    # adjust step size to maintain target rejection rate
    if hmc_ave_rej < hmc_target_ave_rej:
        hmc_step = min(hmc_step*1.01,0.25)
    else:
        hmc_step = max(hmc_step*0.99,.001)
    return hmc_step, hmc_ave_rej

def train_mHPoT():
    ''' mHPoT trainer: sweeps over the training set to update the parameters.
    For each batch of samples compute derivatives w.r.t. parameters
    at the training samples and at the negative samples drawn by HMC sampler.
    Training using either Contrastive Divergence or Fast Persistent Contrastive
    Divergence.'''

    config = ConfigParser()
    config.read('input_configuration')

    verbose = config.getint('VERBOSITY','verbose')

    num_epochs = config.getint('MAIN_PARAMETER_SETTING','num_epochs') # number of sweeps through training set
    batch_size = config.getint('MAIN_PARAMETER_SETTING','batch_size') # number of samples in each batch
    startFH = config.getint('MAIN_PARAMETER_SETTING','startFH') # number of epochs after which FH matrix is updated
    startwd = config.getint('MAIN_PARAMETER_SETTING','startwd') # number of epochs after which parameters are decayed to 0
    doFPCD = config.getint('MAIN_PARAMETER_SETTING','doFPCD') # flag to decide whether to use Fast PCD or CD1
    apply_mask = 0	

    # model parameters
    num_fac = config.getint('MODEL_PARAMETER_SETTING','num_fac') # number of filters in VF matrix
    num_hid_cov =  config.getint('MODEL_PARAMETER_SETTING','num_hid_cov') # number of covariance hidden units
    num_hid_mean =  config.getint('MODEL_PARAMETER_SETTING','num_hid_mean') # number of mean hidden units

    # load data
    data_file_name =  config.get('DATA','data_file_name')
    d = loadmat(data_file_name) # input in the format PxD (P vectorized samples with D dimensions)
    pca_file_name =  config.get('DATA','pca_file_name')
    pca = loadmat(pca_file_name) # input in the format VxD
    totnumcases = d["patches"].shape[0] # number of training samples
    d = d["patches"][0:floor(totnumcases/batch_size)*batch_size,:].copy() 
    totnumcases = d.shape[0]
    num_vis =   pca["pcatransf"].shape[0] # dimensionality of the input (after PCA projection)
    num_batches = int(totnumcases/batch_size) # number of batches
    dev_dat = cmt.CUDAMatrix(np.dot(pca["pcatransf"],d.T)) # VxP, P is number of samples
    save_dir =  config.get('BACK-UP','save_dir')

    # training parameters
    epsilon = config.getfloat('OPTIMIZER_PARAMETERS','epsilon') # overall learning rate for stochastic gradient descent
    epsilonVF = 1.*epsilon # learning rate for VF matrix
    epsilonFH = 0.01*epsilon # learning rate for FH matrix
    epsilonb = 0.01*epsilon # learning rate for scaling parameter coeff
    epsilonw_mean = 0.2*epsilon # learning rate for mean parameters w_mean
    epsilonb_mean = 0.1*epsilon # learning rate for bias_mean
    weightcost_final =  config.getfloat('OPTIMIZER_PARAMETERS','weightcost_final') # weight decay

    # HMC setting
    hmc_step_nr = config.getint('HMC_PARAMETERS','hmc_step_nr') # number of leap-frog steps
    hmc_step =  0.01 # initial step size
    hmc_target_ave_rej =  config.getfloat('HMC_PARAMETERS','hmc_target_ave_rej') # target rejection rate of HMC samples
    hmc_ave_rej =  hmc_target_ave_rej # initial value of running average of rejection rate

    # initialize parameters
    VF = cmt.CUDAMatrix( 0.02 * np.random.randn(num_vis, num_fac)) # VxH
    FH = cmt.CUDAMatrix( np.eye(num_fac,num_hid_cov)) # HxO
    coeff = cmt.CUDAMatrix( np.ones((num_hid_cov, 1)))
    bias_vis = cmt.CUDAMatrix( np.zeros((num_vis, 1)))
    w_mean = cmt.CUDAMatrix( 0.05 * np.random.randn(num_vis, num_hid_mean)) # VxH
    bias_mean = cmt.CUDAMatrix( -1.0*np.ones((num_hid_mean,1)))
    # auxiliary parameters for Fast PCD, initialized at 0
    VFf = cmt.CUDAMatrix(np.zeros((num_vis, num_fac))) # VxH
    VFt = cmt.CUDAMatrix(np.zeros((num_vis, num_fac))) # VxH
    w_meanf = cmt.CUDAMatrix(np.zeros((num_vis, num_hid_mean))) # VxH
    bias_meanf = cmt.CUDAMatrix(np.zeros((num_hid_mean,1)))
    w_meant = cmt.CUDAMatrix(np.zeros((num_vis, num_hid_mean))) # VxH
    bias_meant = cmt.CUDAMatrix(np.zeros((num_hid_mean,1)))

    # initialize variables to store derivatives 
    VFinc = cmt.CUDAMatrix(np.zeros((num_vis, num_fac)))
    FHinc = cmt.CUDAMatrix(np.zeros((num_fac, num_hid_cov)))
    coeffinc = cmt.CUDAMatrix(np.zeros((num_hid_cov, 1)))
    bias_visinc = cmt.CUDAMatrix(np.zeros((num_vis, 1)))
    w_meaninc = cmt.CUDAMatrix(np.zeros((num_vis, num_hid_mean)))
    bias_meaninc = cmt.CUDAMatrix(np.zeros((num_hid_mean, 1)))

    # initialize temporary storage variables
    data = cmt.CUDAMatrix(np.empty((num_vis, batch_size))) # VxP
    negdataini = cmt.CUDAMatrix(np.empty((num_vis, batch_size))) # VxP
    feat = cmt.CUDAMatrix(np.empty((num_fac, batch_size)))
    featsq = cmt.CUDAMatrix(np.empty((num_fac, batch_size)))
    negdata = cmt.CUDAMatrix(np.random.randn(num_vis, batch_size))
    old_energy = cmt.CUDAMatrix(np.zeros((1, batch_size)))
    new_energy = cmt.CUDAMatrix(np.zeros((1, batch_size)))
    gradient = cmt.CUDAMatrix(np.empty((num_vis, batch_size))) # VxP
    thresh = cmt.CUDAMatrix(np.zeros((1, batch_size)))
    feat_mean = cmt.CUDAMatrix(np.empty((num_hid_mean, batch_size)))
    vel = cmt.CUDAMatrix(np.random.randn(num_vis, batch_size))
    normVF = 1    
    
    # other temporary vars
    t1 = cmt.CUDAMatrix(np.empty((num_hid_cov, batch_size)))
    t2 = cmt.CUDAMatrix(np.empty((num_hid_cov, batch_size)))
    t3 = cmt.CUDAMatrix(np.empty((num_fac, batch_size)))
    t4 = cmt.CUDAMatrix(np.empty((1,batch_size)))
    t5 = cmt.CUDAMatrix(np.empty((1,1)))
    t6 = cmt.CUDAMatrix(np.empty((num_vis, batch_size)))
    t7 = cmt.CUDAMatrix(np.empty((num_vis, batch_size)))
    t8 = cmt.CUDAMatrix(np.empty((num_vis, num_fac)))
    t9 = cmt.CUDAMatrix(np.zeros((num_fac, num_hid_cov)))
    t10 = cmt.CUDAMatrix(np.empty((1,num_fac)))
    t11 = cmt.CUDAMatrix(np.empty((1,num_hid_cov)))
    t12 = cmt.CUDAMatrix(np.empty((num_hid_cov, 1)))
    tt0 = cmt.CUDAMatrix(15.*np.ones((num_hid_mean, batch_size)))
    tt1 = cmt.CUDAMatrix(np.empty((num_hid_mean, batch_size)))
    tt2 = cmt.CUDAMatrix(np.empty((num_hid_mean, batch_size)))

    # start training
    for epoch in range(num_epochs): # loop over all training samples

        print "Epoch " + str(epoch + 1)
    
        # anneal current learning rates
        epsilonVFc    = epsilonVF/max(1.,epoch/10.)
        epsilonFHc    = epsilonFH/max(1.,epoch/10.)
        epsilonbc    = epsilonb/max(1.,epoch/10.)
        epsilonw_meanc = epsilonw_mean/max(1.,epoch/10.)
        epsilonb_meanc = epsilonb_mean/max(1.,epoch/10.)
        weightcost = weightcost_final

        if epoch <= startFH:
            epsilonFHc = 0 
        if epoch <= startwd:	
            weightcost = 0

        for batch in range(num_batches): # loop over batches

            # get current minibatch
            data = dev_dat.slice(batch*batch_size,(batch + 1)*batch_size) # VxP (nr dims x nr samples)

            cmt.dot(VF.T, data, target = feat) # HxV VxP -> HxP
            feat.mult(feat, target = featsq)   # HxP
            cmt.dot(FH.T,featsq, target = t1) # OxP, t1=(VF*x)^2*FH 
            t1.mult(0.5)
            t1.add(1.) # OxP 1+.5(VF*x)^2*FH
            t1.reciprocal(target = t2) # OxP
            t2.mult_by_col(coeff) # OxP coeff/(1+.5(VF*x)^2*FH)            
            # positive sample increments
            cmt.dot(featsq,t2.T,target = FHinc) # HxO 
            cmt.dot(FH,t2, target = t3) # HxP
            t3.mult(feat) # HxP
            cmt.dot(data,t3.T, target = VFinc) # VxH
            cmt.log(t1) # OxP
            t1.sum(axis = 1, target = coeffinc) # Ox1
            # visible bias
            data.sum(axis = 1, target = bias_visinc)
            bias_visinc.mult(-1.)
            # mean part
            cmt.dot(w_mean.T, data, target = feat_mean) # HxP (nr mean hiddens x nr samples)
            feat_mean.add_col_vec(bias_mean) # HxP
            feat_mean.apply_sigmoid() # HxP
            feat_mean.mult(-1.)
            cmt.dot(data, feat_mean.T, target = w_meaninc)
            feat_mean.sum(axis = 1, target = bias_meaninc)
               

            
            # update total weight vector
            if doFPCD == 1:
                VF.add(VFf,target = VFt)
                w_mean.add(w_meanf,target = w_meant)
                bias_mean.add(bias_meanf, target = bias_meant)
            else:
                VFt.assign(VF)
                w_meant.assign(w_mean)
                bias_meant.assign(bias_mean)
            
            # HMC sampling: draw an approximate sample from the model (run for only one MC step)
            if doFPCD == 1:
	        negdataini.assign(negdata)
	    else:
		negdataini.assign(data)
            hmc_step, hmc_ave_rej = draw_HMC_samples(negdataini,negdata,vel,gradient,new_energy,old_energy,VFt,FH,coeff,bias_vis,w_meant,bias_meant,hmc_step,hmc_step_nr,hmc_ave_rej,hmc_target_ave_rej,t1,t2,t3,t4,t5,t6,t7,tt0,tt1,tt2,thresh,feat,featsq,batch_size,feat_mean)
            


            # compute derivatives at the negative samples
            cmt.dot(VFt.T, negdata, target = feat) # HxP 
            feat.mult(feat, target = featsq)   # HxP
            cmt.dot(FH.T,featsq, target = t1) # OxP, t1=(VF*x)^2*FH 
            t1.mult(0.5)
            t1.add(1.) # OxP 1+.5(VF*x)^2*FH
            t1.reciprocal(target = t2) # OxP
            t2.mult_by_col(coeff) # OxP coeff/(1+.5(VF*x)^2*FH)            
            # positive sample increments
            FHinc.subtract_dot(featsq,t2.T)
            cmt.dot(FH,t2, target = t3) # HxP
            t3.mult(feat) # HxP
            VFinc.subtract_dot(negdata,t3.T)
            cmt.log(t1) # OxP
            t1.mult(-1.)
            coeffinc.add_sums(t1, axis = 1)
            # visible bias
            bias_visinc.add_sums(negdata, axis = 1)
            # mean part
            cmt.dot(w_meant.T, negdata, target = feat_mean) # HxP 
            feat_mean.add_col_vec(bias_meant) # HxP
            feat_mean.apply_sigmoid() # HxP
            w_meaninc.add_dot(negdata, feat_mean.T)
            bias_meaninc.add_sums(feat_mean, axis = 1)


            # update parameters
            if doFPCD == 1:
                # fast
                VFf.mult(19./20) # decay to 0
                VFf.add_mult(VFinc, -epsilonVF/batch_size) # raise energy at negative particles
                w_meanf.mult(19./20) # decay to 0
                w_meanf.add_mult(w_meaninc, -epsilonw_mean/batch_size)  # raise energy at negative particles
                bias_meanf.mult(19./20) # decay to 0
                bias_meanf.add_mult(bias_meaninc, -epsilonb_mean/batch_size) # raise energy at negative particles
            # regular
            VFinc.add_mult(VF, weightcost) # L2 decay
            VF.add_mult(VFinc, -epsilonVFc/batch_size) # update
            # normalize columns of VF: normalize by running average of their norm 
            VF.mult(VF, target = t8)
            t8.sum(axis = 0, target = t10)
            cmt.sqrt(t10)
            t10.sum(axis=1,target = t5)
            t5.copy_to_host()
            normVF = .95*normVF + (.05/num_fac) * t5.numpy_array[0,0] # estimate norm
            t10.reciprocal()
            VF.mult_by_row(t10) 
            VF.mult(normVF) 
            coeff.add_mult(coeffinc, -epsilonbc/batch_size)            
            # make sure coeff does not go negative (actually below .1)
            coeff.greater_than(.1,target = t12)
            coeff.mult(t12)
            t12.mult(-.1)
            t12.add(.1)
            coeff.add(t12)
            # average scale parameters
            coeff.sum(axis=0, target = t5)
            t5.mult(1./num_hid_cov)
            coeff.mult(0.)
            coeff.add_row_vec(t5)
            # update visible biases
            bias_vis.add_mult(bias_visinc, -epsilonbc/batch_size)

            if epoch > startFH:
                FHinc.add_mult(FH.sign(), weightcost) # L1 regularization
       		FH.add_mult(FHinc, -epsilonFHc/batch_size) # update
	        # set to 0 negative entries in FH
        	FH.greater_than(0, target = t9)
	        FH.mult(t9)
                if apply_mask==1:
                    FH.mult(mask) # this is useful to enforce a topography
		# normalize columns of FH: L1 norm set to 1 in each column
		FH.sum(axis = 0, target = t11)               
		t11.reciprocal()
		FH.mult_by_row(t11) 
            # update mean parameters
            w_meaninc.add_mult(w_mean,weightcost) # L2 weight decay
            w_mean.add_mult(w_meaninc, -epsilonw_meanc/batch_size)
            bias_mean.add_mult(bias_meaninc, -epsilonb_meanc/batch_size)


        if verbose == 1: # print some info
            print "VF: " + '%3.2e' % VF.euclid_norm() + ", DVF: " + '%3.2e' % (VFinc.euclid_norm()*(epsilonVFc/batch_size)) + ", FH: " + '%3.2e' % FH.euclid_norm() + ", DFH: " + '%3.2e' % (FHinc.euclid_norm()*(epsilonFHc/batch_size)) + ", coeff: " + '%3.2e' % coeff.euclid_norm() + ", Dcoeff: " + '%3.2e' % (coeffinc.euclid_norm()*(epsilonbc/batch_size)) + ", bias_vis: " + '%3.2e' % bias_vis.euclid_norm() + ", Dbias_vis: " + '%3.2e' % (bias_visinc.euclid_norm()*(epsilonbc/batch_size)) + ", wm: " + '%3.2e' % w_mean.euclid_norm() + ", Dwm: " + '%3.2e' % (w_meaninc.euclid_norm()*(epsilonw_meanc/batch_size)) + ", bm: " + '%3.2e' % bias_mean.euclid_norm() + ", Dbm: " + '%3.2e' % (bias_meaninc.euclid_norm()*(epsilonb_meanc/batch_size)) + ", step: " + '%3.2e' % hmc_step  +  ", rej: " + '%3.2e' % hmc_ave_rej 
            sys.stdout.flush()
        # back-up every once in a while 
        if np.mod(epoch,10) == 0:
            VF.copy_to_host()
            FH.copy_to_host()
            coeff.copy_to_host()
            w_mean.copy_to_host()
            bias_mean.copy_to_host()
            bias_vis.copy_to_host()
            data.copy_to_host()
            negdata.copy_to_host()
            savemat(save_dir + "params_temp", {'VF':VF.numpy_array,'FH':FH.numpy_array,'coeff': coeff.numpy_array, 'bias_vis': bias_vis.numpy_array,'w_mean': w_mean.numpy_array, 'bias_mean': bias_mean.numpy_array,'data':data.numpy_array,'negdata':negdata.numpy_array,'epoch':epoch})    
    # final back-up
    VF.copy_to_host()
    FH.copy_to_host()
    coeff.copy_to_host()
    bias_vis.copy_to_host()
    w_mean.copy_to_host()
    bias_mean.copy_to_host()
    savemat(save_dir + "params_fac" + str(num_fac) + "_cov" + str(num_hid_cov) + "_mean" + str(num_hid_mean), {'VF':VF.numpy_array,'FH':FH.numpy_array,'coeff': coeff.numpy_array, 'bias_vis': bias_vis.numpy_array, 'w_mean': w_mean.numpy_array, 'bias_mean': bias_mean.numpy_array, 'epoch':epoch})



if __name__ == "__main__":
  # initialize CUDA
  cmt.cuda_set_device(gpu_lock.obtain_lock_id()) # comment if you do not have a locking system to choose the GPU board number
  cmt.cublas_init()
  cmt.CUDAMatrix.init_random(1)
  train_mHPoT()
  cmt.cublas_shutdown()


